// This may look like C code, but it is really -*- C++ -*-
// RTSS --- Railway Total System Simulator
// (c) TAKAGI Ryo (at Kogakuin University)
// RS_XmlReader.hh --- class RS_XmlReader and related classes
// -----



#ifndef RS_XmlReader_HH
#define RS_XmlReader_HH

#include <cstdlib>
#include <string>
#include <map>
#include <iostream>
#include "RS_Object.hh"
#include "tinyxml.h"

using std :: map ;
using std :: string ;



/**************************************************************************
 * string getClassString ( const TiXmlNode * _n_in ) : Non-member function
 *  to find the class string.
 * _n_in : tag containing attribute "class".
 * 
 * string getNameString ( const TiXmlNode * _n_in , bool _sw = false ) :
 *  Non-member function to find the name string.
 * _n_in : tag containing attribute "name".
 * _sw : true if attributes other than "name" is rejected.
 * 
 * string getStringAttribute ( string const & _nx , TiXmlNode const * _n_in ,
 * bool _sw = false , bool _sx = false ) :
 *  Non-member function to find the name string.
 * _nx : name of the attribute.
 * _n_in : tag containing attribute _nx.
 * _sw : true if attributes other than _nx is rejected.
 * _sx : true if this attribute is optional.
 **************************************************************************/
string getClassString ( TiXmlNode const * ) ;
string getNameString ( TiXmlNode const * , bool = false ) ;
string getStringAttribute
( string const & , TiXmlNode const * , bool = false , bool = false ) ;

/**************************************************************************
 * void getAttributeMap ( TiXmlNode const * _n_in , map < string , sring >
 *   & _a_m ) : Non-member function to create an attribute map.
 * _n_in : tag to be searched.
 * _a_m : map to be created.
 *
 * map < string , string > :: iterator findAttributeInMap ( map < string ,
 *   string > & _a_m , string const & _a_n , string const & _et = "" , bool
 *   _swreq = false ) : Non-member function to find an iterator pointing to
 *   the attribute.
 * _a_m : map of the attributes created by getAttributeMap().
 * _a_n : name of the attribute to be searched.
 * _et : In error messages "Error: xxxxx: ....", string xxxxx will be
 *   replaced by _et.
 * _swreq : true if this attribute is required.
 **************************************************************************/
void getAttributeMap ( TiXmlNode const * , map < string , string > & ) ;
map < string , string > :: iterator findAttributeInMap
( map < string , string > & , string const & ,
  string const & = "in RS_XR fAIM" , bool = false ) ;




/************************************************************************
 * class RS_XmlReader
 * ----
 * This class reads XML-format data input files and create objects.
 * このクラスは，XML形式の入力ファイルを読み取り，オブジェクトを生成する。
 ************************************************************************/
class RS_XmlReader
  : public virtual RS_Object ,
    public map < string , RS_ObjectContainer * >
{
  // Information on current files.
  bool has_open_document ;
  string current_document_file_descriptor ;
  string current_drive ;
  string current_directory ;
  TiXmlDocument * curdoc ;
  TiXmlNode const * curnode ;
  TiXmlNode const * dtag ;

  // Subfunctions of processElement() member function.
  bool processCreateTag () ;

  explicit RS_XmlReader ( RS_XmlReader & ) ;

public :
  // Constructor.
  explicit RS_XmlReader ()
    : map < string , RS_ObjectContainer * > () , has_open_document ( false ) ,
      current_document_file_descriptor ( "" ) , current_drive ( "" ) ,
      current_directory ( "" ) , curdoc ( 0 ) , curnode ( 0 ) , dtag ( 0 ) {}
  
  /**********************************************************************
   * bool openXmlDocument ( const string & filedcr ,
   *  const string & curdrive , const string & curdir )
   * Opens an XML file and locate the first tag to be processed.
   * Returns true if successful.
   **********************************************************************/
  bool openXmlDocument
  ( string const & , string const & , string const & ) ;

  /**********************************************************************
   * bool hasNextElement ()
   * Locate next unprocessed tag. Returns true if there is one and the
   * operation successful. If no unprocessed tag is found, close the
   * current document.
   **********************************************************************/
  bool hasNextElement () ;

  /**********************************************************************
   * bool processElement ()
   * Process the current element. Returns true if successful.
   **********************************************************************/
  bool processElement () ;
} ;




// -----
// Variables for PCRE.
// -----
const int RP_OVECCOUNT = 30 ;
const int RP_PATHMAX = 16384 ;




// -----
// Path string analysis and manipulation class.
// -----
class RTSS_Path_String
  : public string
{
private :
  int * ovector ;

public :

  // -----
  // Constructors
  // コンストラクタ
  // -----
  RTSS_Path_String () ;
  RTSS_Path_String ( const char * ) ;
  RTSS_Path_String ( const string & ) ;
  RTSS_Path_String ( const RTSS_Path_String & ) ;

  // -----
  // Substitution operator
  // 代入演算子
  // -----
  RTSS_Path_String & operator= ( const RTSS_Path_String & ) ;

  // -----
  // Destructor
  // デストラクタ
  // -----
  ~RTSS_Path_String () { delete [] ovector ; }

  // -----
  // Returns true if the full path name is of Windows style. If true, the
  // drive name will be set to the argument string variable, and the path
  // name in the class will be truncated.
  // ウィンドウズ・スタイルのフルパス名である場合 true を返す. True なら，
  // そのさい string 引数にドライブ名がセットされ，クラスに格納されている
  // ファイル名はドライブ名を取り去られる.
  // -----
  bool isWindowsStyleFullPath ( string & ) ;

  // -----
  // Returns true if the name has .xml extension.
  // .xml 拡張子を持つ場合 true を返す.
  // -----
  bool hasXmlExtension () const ;

  // -----
  // Returns true if the name has .ods extension.
  // .ods 拡張子を持つ場合 true を返す.
  // -----
  bool hasOdsExtension () const ;

  // -----
  // Returns true if the name has .csv extension.
  // .ods 拡張子を持つ場合 true を返す.
  // -----
  bool hasCsvExtension () const ;

  // -----
  // Process Path. Arguments: whether it is windows-style (bool), current
  // drive (const string &), current directory (const string &), resulting
  // file drive name (string &), resulting file directory name (string &),
  // resulting file name (string &). Returns true if successful.
  // パス名を処理する. 引数: ウィンドウズ・スタイルか否か (bool), カレント
  // ドライブ (const string &), カレントディレクトリ (const string &),
  // 結果的なファイルのドライブ名 (string &), 同ディレクトリ名 (string &),
  // 同ファイル名 (string &). 成功すれば true.
  // -----
  bool processPath ( bool , const string & , const string & ,
                     string & , string & , string & ) ;
} ;


#endif	// RS_XmlReader_HH
